/*******************************************************************
*
*                       Cinepak for SATURN Player 
*                                by SOJ
*                         usage sample program
*
*		Sample to play a movie file on a CD using a file system 
		Playback is at 15-bit RGB using VDP1
* 		The movie file name is "SAMPLE.CPK" 
*
*                      Copyright(c) 1994,1995 SEGA
*
*   Comment: main module
*   File   : SMPCPK4.c
*   Date   : 1994-10-31
*   Author : H.G
*
*******************************************************************/
/* Play a movie of uncompressed images */
/* #define NONE_COMPRESS_VIDEO */

/* Expand to HWRAM once and transfer to VRAM by DMA  */
#define OUTPUT_WORK_RAM

/* Define this when you want the Cinepac library to use a slave CPU.B */
/* #define DUAL_CPU */

/* Set the playback start trigger size. */
/*#define START_TRG_SIZE		 (RING_BUF_SIZ - 24*2048) */
#define START_TRG_SIZE		 (30 * 2048)

#include <machine.h>
#include <string.h>
#define _SH
#include "sega_xpt.h"
#include "sega_sys.h"
#include "sega_def.h"
#include "sega_mth.h"
#include "sega_scl.h" 
#include "sega_int.h"
#define  _SPR2_
#include "sega_spr.h"
#include "sega_cdc.h"
#include "sega_gfs.h"
#include "sega_dma.h"
#include "sega_per.h"
#include "sega_snd.h"
#include "sega_cpk.h"



static void smpVblIn(void);
static void smpVblOut(void);
static void smpSprEnd(void);


/*------------------------- s}N萔t -------------------------*/

/* VDP1 VRAM Address */
#define ADDR_VDP1 				(0x25C00000)

/* VRAM transfer destination address  */
#define ADDR_VRAM_CPK 			(0x25C08000)
#define ADDR_VRAM_MON 			(0x25C30000)

/* PCM RAM transfer address and number of samples */
#define	PCM_ADDR				((void*)0x25a20000)
#define	PCM_SIZE				(4096L*16)

#define	RING_BUF_SIZ	(1024L*400)

/* TV screen size */
#define   DISP_XSIZE       320
#define   DISP_YSIZE       224

/* Sprite Variables */
#define SPRITE_CPK		(3)
#define LUX_CPK			(0)
#define LUY_CPK			(0)
#define DX_CPK			(320)
#define DY_CPK			(224)

/* Number of movies to play */
#define	FILE_NUM		(2)

/* Minimum and maximum value of the volume */
#define LEVEL_MIN	(0)
#define LEVEL_MAX	(7)

/* Panning Min, Max, and Center */
#define PAN_MIN		(0)					/* Pan Left: maximum on the left, zero on the right */
#define PAN_MAX		(31)				/* Pan Right: right is maximum, left is zero */
#define PAN_CENTER	((PAN_MAX + 1) / 2)	/* Center: maximum left and right  	*/


/*----------------------- s Global variables t -----------------------*/

/* Work Buffer */
static Uint32 g_movie_work[CPK_15WORK_DSIZE];

/* Ring Buffer */
static Uint32 g_movie_buf[RING_BUF_SIZ / sizeof(Uint32)];

/* Name of the file to play */
static char *filename = "SAMPLE.CPK";

/* Flag to end sprite drawing  TRUE : Stop Drawing */
static volatile Bool g_spr_end;

/* Workram buffer to output once  */
#ifdef OUTPUT_WORK_RAM
	Uint32 work_ram_buf[320L * 224L / 2];
	Uint32 cp_size;
#endif

/* Game Pad Info */
static Uint32	PadWorkArea[7];
static Uint16	PadData1  = 0x0000;
static Uint16	PadData1E = 0x0000;
static Uint16	PadData2  = 0x0000;
static Uint16	PadData2E = 0x0000;

/* Pause State */
static Bool		g_pause_at_once;
static Bool		g_pause_keyframe;

/* Volume Level */
Sint32 g_level = LEVEL_MAX;

/* Panning Value, Left = 0, Right = 31. */
Sint32 g_pan = PAN_CENTER;

/* Pan value: value specified as an argument */
Sint32 g_pan_tbl[PAN_MAX + 1] = {
	0x1f, 0x1e, 0x1d, 0x1c, 0x1b, 0x1a, 0x19, 0x18, 
	0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f
};

/********************************************************************/
/* Function to be called when an error occurs.	    				*/
/********************************************************************/
void errGfsFunc(void *obj, Sint32 ec)
{
	/* Error Handling */
}

void errCpkFunc(void *obj, Sint32 ec)
{
	/* Error Handling */
}


/* Switching the frame buffer and waiting for the drawing to finish  */
void smp_WaitDisplayFrame(void)
{
	/* Waiting for (previous) drawing from VRAM to framebuffer  */
	while (g_spr_end == FALSE) ;

	/* Framebuffer switching  */
	SCL_DisplayFrame();

	g_spr_end = FALSE;
}
void smp_DisplayFrameWait(void)
{
	/* Framebuffer switching  */
	SCL_DisplayFrame();

	g_spr_end = FALSE;

	/* Waiting for (previous) drawing from VRAM to framebuffer  */
	while (g_spr_end == FALSE) ;

}


/*====================== Screen Display Processing ===========================*/
#define CHARADDR(addr)		((Uint16)(((addr) >> 3) & 0x0000ffff))
#define CHARSIZE(dx, dy)	((Uint16)((((dx) >> 3)<<8) | (dy)))

static SprSpCmd SpriteCmd[] = {
	{
		/* control  */ (JUMP_NEXT | FUNC_SCLIP),
		/* link     */  0,
		/* drawMode */  0,
		/* color    */  0,
		/* charAddr */  0,
		/* charSize */  0,
		/* ax, ay   */  0, 0,
		/* bx, by   */  0, 0,
		/* cx, cy   */  351, 223,
		/* dx, dy   */  0, 0,
		/* grshAddr */  0,
		/* dummy    */  0
	}, {
		/* control  */ (JUMP_NEXT | FUNC_LCOORD),
		/* link     */  0,
		/* drawMode */  0,
		/* color    */  0,
		/* charAddr */  0,
		/* charSize */  0,
		/* ax, ay   */  0, 0,
		/* bx, by   */  0, 0,
		/* cx, cy   */  0, 0,
		/* dx, dy   */  0, 0,
		/* grshAddr */  0,
		/* dummy    */  0
	}, {
		/* control  */ (JUMP_NEXT | FUNC_UCLIP),
		/* link     */  0,
		/* drawMode */  0,
		/* color    */  0,
		/* charAddr */  0,
		/* charSize */  0,
		/* ax, ay   */  0, 0,
		/* bx, by   */  0, 0,
		/* cx, cy   */  351, 223,
		/* dx, dy   */  0, 0,
		/* grshAddr */  0,
		/* dummy    */  0
	}, {
		/* control  */  (JUMP_NEXT | FUNC_DISTORSP),
		/* link     */  0,
		/* drawMode */  0x04A8,
		/* color    */  0x0000,		/*  */
		/* charAddr */  CHARADDR(ADDR_VRAM_CPK),
		/* charSize */  CHARSIZE(DX_CPK, DY_CPK),
		/* ax, ay   */  LUX_CPK, 				LUY_CPK,
		/* bx, by   */  LUX_CPK + DX_CPK - 1, 	LUY_CPK,
		/* cx, cy   */  LUX_CPK + DX_CPK - 1, 	LUY_CPK + DY_CPK - 1,
		/* dx, dy   */  LUX_CPK, 				LUY_CPK + DY_CPK - 1,
		/* grshAddr */  0,
		/* dummy    */  0
	}, {
		/* [ 0]     */
		/* control  */ 	(CTRL_END),
		/* link     */ 	0,
		/* drawMode */ 	0,
		/* color    */ 	0,
		/* charAddr */ 	0,
		/* charSize */ 	0,
		/* ax, ay   */  0,   0,
		/* bx, by   */  0,   0,
		/* cx, cy   */  0,   0,
		/* dx, dy   */  0,   0,
		/* rshAddr */   0,
		/* dummy    */  0
	}
};

void eraseVram(void)
{
	Sint32 *addr = (Sint32 *)(0x25C80000);	/* end of VDP1 VRAM */

	do {
		*--addr = 0x80008000;
	} while ((Sint32)addr > 0x25C00000);	/* start of VDP1 VRAM */
}

static void dispInit(void)
{
    Uint8  *VRAM;
	Uint16			BackCol;

    set_imask(0);
    SCL_Vdp2Init();
    SCL_SetPriority(SCL_SP0|SCL_SP1|SCL_SP2|SCL_SP3|
                    SCL_SP4|SCL_SP5|SCL_SP6|SCL_SP7,7);
    SCL_SetSpriteMode(SCL_TYPE1,SCL_MIX,SCL_SP_WINDOW);

	/* Vblank setting  */
	INT_ChgMsk(INT_MSK_NULL,INT_MSK_VBLK_IN | INT_MSK_VBLK_OUT);

	/* V_Blank Out (Wait for an interrupt ) */
	(*((volatile Uint32 *)0x25fe00a4)) &= 0xfffffffc;	/* ܂NA */
	while( !((*((volatile Uint32 *)0x25fe00a4)) & 2) );
	PER_LInit(PER_KD_PERTIM,2,PER_SIZE_DGT,PadWorkArea,0);

	INT_SetScuFunc(INT_SCU_VBLK_IN, smpVblIn);
	INT_SetScuFunc(INT_SCU_VBLK_OUT, smpVblOut);
	INT_ChgMsk(INT_MSK_VBLK_IN | INT_MSK_VBLK_OUT,INT_MSK_NULL);

	/* Sprite drawing end interrupt setting  */
	INT_ChgMsk(INT_MSK_NULL,INT_MSK_SPR);
	INT_SetScuFunc(INT_SCU_SPR, smpSprEnd);
	INT_ChgMsk(INT_MSK_SPR, INT_MSK_NULL);

	BackCol = 0x0000;
	SCL_SetBack(SCL_VDP2_VRAM+0x80000-2,1,&BackCol);

    SPR_Initial(&VRAM);
    SPR_SetEraseData(RGB16_COLOR(0,0,0),0,0,DISP_XSIZE-1,DISP_YSIZE-1);
	SCL_SetFrameInterval(1);
	SCL_DisplayFrame();
	SCL_DisplayFrame();
	SCL_SetFrameInterval(-1);

	eraseVram();
}

static void setSprite(Sint32 sizeX, Sint32 sizeY)
{
	Uint32 		luX, luY, rdX, rdY;

	SpriteCmd[SPRITE_CPK].charSize = CHARSIZE(sizeX, sizeY);
	luX = (320 - sizeX) / 2;
	luY = (224 - sizeY) / 2;
	rdX = luX + sizeX - 1;
	rdY = luY + sizeY - 1;
	SpriteCmd[SPRITE_CPK].ax = luX;
	SpriteCmd[SPRITE_CPK].ay = luY;
	SpriteCmd[SPRITE_CPK].bx = rdX;
	SpriteCmd[SPRITE_CPK].by = luY;
	SpriteCmd[SPRITE_CPK].cx = rdX;
	SpriteCmd[SPRITE_CPK].cy = rdY;
	SpriteCmd[SPRITE_CPK].dx = luX;
	SpriteCmd[SPRITE_CPK].dy = rdY;

	memcpy((void *)ADDR_VDP1, SpriteCmd, sizeof(SpriteCmd));
}


/*====================== V-blank processing ===========================*/
static void smpVblIn(void)
{
	/* Call Cinepak's VblIn routine  */
	CPK_VblIn();

	/* Must be run to use the graphics library  */
	SCL_VblankStart();

}

static void smpVblOut(void)
{
	PerDgtInfo	*pad;
	PerDgtInfo	*mul_info;

	PER_LGetPer((PerGetPer **)&pad, (PerMulInfo **)&mul_info);
	PadData1  = pad[0].data ^ 0xffff;
	PadData1E |= pad[0].push ^ 0xffff;
	PadData2  = pad[1].data ^ 0xffff;
	PadData2E |= pad[1].push ^ 0xffff;

	/* Must be run to use the graphics library  */
	SCL_VblankEnd();
}

static void smpSprEnd(void)
{
	g_spr_end = TRUE;
}

/*====================== Sound processing ===========================*/
#define SDDRVS_TSK_SIZE			(0x6000)
#define BOOTSND_MAP_SIZE		(0x0100)
Sint32 sddrvs_tsk[SDDRVS_TSK_SIZE / 4];
Sint32 bootsnd_map[BOOTSND_MAP_SIZE / 4];

static Sint32 fileLoad(Sint8 *name, void *addr, Sint32 bsize)
{
	Sint32 		fid, i;

	for (i = 0; i < 10; i++) {
		fid = GFS_NameToId(name);
		if (fid >= 0) {
			GFS_Load(fid, 0, addr, bsize);
			return 0;
		}
	}
	return -1;
}

static void sndInit(void)
{
	SndIniDt 	snd_init;

	if (fileLoad("SDDRVS.TSK", (void *)sddrvs_tsk, SDDRVS_TSK_SIZE)) {
		return;
	}
	if (fileLoad("BOOTSND.MAP", (void *)bootsnd_map, BOOTSND_MAP_SIZE)) {
		return;
	}
	SND_INI_PRG_ADR(snd_init) 	= (Uint16 *)sddrvs_tsk;
	SND_INI_PRG_SZ(snd_init) 	= (Uint16 )SDDRVS_TSK_SIZE;
	SND_INI_ARA_ADR(snd_init) 	= (Uint16 *)bootsnd_map;
	SND_INI_ARA_SZ(snd_init) 	= (Uint16)BOOTSND_MAP_SIZE;
	SND_Init(&snd_init);
	SND_ChgMap(0);
}


/*====================== File processing ===========================*/
/* Maximum number of files in the root directory */
#define MAX_DIR		100

/* Maximum number of files to open at the same time */
#define OPEN_MAX	5

/* Directory information management area */
static GfsDirTbl	dir_tbl;

/* Information including file name */
static GfsDirName dir_name[MAX_DIR];

/* GFS work area */
static Uint8 gfs_work[GFS_WORK_SIZE(OPEN_MAX)];


static void fileInit(void)
{
	Sint32 file_num;

	/* GFS initialization  */
	GFS_DIRTBL_TYPE(&dir_tbl) = GFS_DIR_NAME;
	GFS_DIRTBL_DIRNAME(&dir_tbl) = dir_name;
	GFS_DIRTBL_NDIR(&dir_tbl) = MAX_DIR;

	/* File system initialization  */
	file_num = GFS_Init(OPEN_MAX, gfs_work, &dir_tbl);
	if (file_num < 0) {
		return;
	}

	/* Error function settings  */
	GFS_SetErrFunc(errGfsFunc, NULL);
}

static GfsHn fileOpen(char *fname)
{
    Sint32 fid;

    /* Find the file identifier from the file name  */
    fid = GFS_NameToId(fname);
    return GFS_Open(fid);
}

static void fileClose(GfsHn fp)
{
	GFS_Close(fp);
}


/* Cinepak initialization */
void cpkInit(void)
{
	/* Cinepak initialization */
	CPK_Init();

	/* Error function settings */
	CPK_SetErrFunc(errCpkFunc, NULL);

#ifdef DUAL_CPU
	/* Let the slave CPU be used by the Cinepac library. */
	CPK_SetCpu(CPK_CPU_DUAL);
#endif
}

/* Generate movie handle */
static CpkHn createMovie(GfsHn gfs)
{
	Uint32 			movie_x, movie_y;
	CpkCreatePara	para;
	CpkHeader		*header;
	CpkHn			cpk;

	/* Generate movie handle */
	CPK_PARA_WORK_ADDR(&para) = g_movie_work;
	CPK_PARA_WORK_SIZE(&para) = CPK_15WORK_BSIZE;
	CPK_PARA_BUF_ADDR(&para) = g_movie_buf;
	CPK_PARA_BUF_SIZE(&para) = RING_BUF_SIZ;
	CPK_PARA_PCM_ADDR(&para) = PCM_ADDR;
	CPK_PARA_PCM_SIZE(&para) = PCM_SIZE;
	cpk = CPK_CreateGfsMovie(&para, gfs);
	if (cpk == NULL) {
		return NULL;
	}

	/* Set the number of display colors to 32000 colors  */
	CPK_SetColor(cpk, CPK_COLOR_15BIT);

	/* Read header */
	/* CPK_Preload and */
	/* CPK_GetHeader does not need to be called if the size of the movie is known in advance */
	/* Playback starts slightly earlier if CPK_PreloadHeader () is not called */ 
	CPK_PreloadHeader(cpk);

	/* Get the size of the movie  */
	header = CPK_GetHeader(cpk);
	movie_x = header->width;
	movie_y = header->height;
	setSprite(movie_x, movie_y);

	/* Set the deployment address of the movie  */
#ifdef OUTPUT_WORK_RAM
	CPK_SetDecodeAddr(cpk, (void *)work_ram_buf, 2 * movie_x);
	cp_size = (Uint32)(movie_x * movie_y * 2);
#else
	CPK_SetDecodeAddr(cpk, (void *)ADDR_VRAM_CPK, 2 * movie_x);
#endif

#ifdef START_TRG_SIZE
	/* Playback start trigger size setting [byte] */
	CPK_SetStartTrgSize(cpk, START_TRG_SIZE);
#endif

	CPK_SetVolume(cpk, g_level);
	CPK_SetPan(cpk, g_pan_tbl[g_pan]);

	return cpk;
}

void smp_PadCtrlCpk(CpkHn cpk, Uint16 PadData1EW, Uint16 PadData1W)
{
	/* Pause */
	if (PadData1EW & PER_DGT_S) {
		PadData1EW = 0;
		if (g_pause_at_once == FALSE) {
			/* Start */

			/* CPK_PAUSE_ON_AT_ONCE Since the processing is heavy, make sure to display a still image * before execution. 
			 * if this is not done results may be unexpected.
			 */
			/* Framebuffer switching and waiting for the end of drawing  */
			smp_DisplayFrameWait();

			g_pause_at_once = TRUE;
			CPK_Pause(cpk, CPK_PAUSE_ON_AT_ONCE);
		} else {
			/* Resume */
			g_pause_at_once = FALSE;
			CPK_Pause(cpk, CPK_PAUSE_OFF);
		}
	}
	if (PadData1EW & PER_DGT_C) {
		PadData1EW = 0;
		if (g_pause_keyframe == FALSE) {
			/* Start */
			g_pause_keyframe = TRUE;
			CPK_Pause(cpk, CPK_PAUSE_ON_KEYFRAME);
		} else {
			/* Resume */
			g_pause_keyframe = FALSE;
			CPK_Pause(cpk, CPK_PAUSE_OFF);
		}
	}

	/* Stop */
	if (PadData1EW & PER_DGT_A) {
		PadData1EW = 0;
		CPK_Stop(cpk);
	}

#if 0
	/* Change */
	if ((g_change_flag == FALSE) && (PadData1EW & PER_DGT_A)) {
		PadData1EW = 0;
		g_change_flag = TRUE;
		goNextStm(stm[1]);
	}
	/* Forced switching of movie  */
	if (g_change_flag == TRUE) {
		if (CPK_CheckChange() == CPK_CHANGE_OK_AT_ONCE) {
			CPK_Change();
			g_change_flag = FALSE;
		}
	}
#endif

	/* Increase Volume */
	if (PadData1EW & PER_DGT_U) {
		PadData1EW = 0;
		g_level = MIN(g_level + 1, LEVEL_MAX);
		CPK_SetVolume(cpk, g_level);
	}
	/* Decrease Volume*/
	if (PadData1EW & PER_DGT_D) {
		PadData1EW = 0;
		g_level = MAX(g_level - 1, LEVEL_MIN);
		CPK_SetVolume(cpk, g_level);
	}

	/* Pan E From the right (maximum on the right, gradually weakening on the left)  */
	if (PadData1EW & PER_DGT_R) {
		PadData1EW = 0;
		g_pan = MIN(g_pan + 1, PAN_MAX);
		CPK_SetPan(cpk, g_pan_tbl[g_pan]);
	}
	/* Pan E From the left (maximum left, gradually weakening right)  */
	if (PadData1EW & PER_DGT_L) {
		PadData1EW = 0;
		g_pan = MAX(g_pan - 1, PAN_MIN);
		CPK_SetPan(cpk, g_pan_tbl[g_pan]);
	}
}

#define SMP_PAD_MODE_NUM		(1)
Sint32 smp_pad_mode = 0;

/* Pad control  */
void smp_PadCtrl(CpkHn cpk)
{
	Uint16   PadData1EW;
	Uint16   PadData1W;

	PadData1EW = PadData1E;
	PadData1E = 0;
	PadData1W = PadData1;
	PadData1 = 0;

	/* Switching the pad processing target  */
	if (PadData1EW & PER_DGT_TL) {
		PadData1EW = 0;
		smp_pad_mode++;
	}

	switch (smp_pad_mode % SMP_PAD_MODE_NUM) {
	case 0:
		/* Pad control: Playback library  */
		smp_PadCtrlCpk(cpk, PadData1EW, PadData1W);
		break;
	}
}

void main(void)
{
	CpkHn 			cpk;
	GfsHn			gfs;
	Uint32			restart;

#ifdef OUTPUT_WORK_RAM
	DMA_ScuInit();
#endif

	/* Variable Initialization */
	g_spr_end = TRUE;

	/* Sprite Initialization */
	dispInit();

	/* File Initialization */
	fileInit();

	/* Sound Initialization */
	sndInit();

	/* Cinepak initialization */
	cpkInit();

	restart = 1;

	while (1) {
		if (restart) {

			/* File open */
			if ((gfs = fileOpen(filename)) == NULL) {
				return;
			}

			/* Generate movie handle */
			cpk = createMovie(gfs);

			/* Movie Start */
			CPK_Start(cpk);

			restart = 0;
		}

		/* Movie playback process  */
		CPK_Task(cpk);

		/* Check screen display request  */
		if (CPK_IsDispTime(cpk) == TRUE) {

#ifdef OUTPUT_WORK_RAM
			DMA_ScuMemCopy((void *)ADDR_VRAM_CPK, 
				(void *)work_ram_buf, cp_size);
#endif

			/* Framebuffer switching and waiting for the end of drawing  */
			smp_DisplayFrameWait();

			/* \̒ʒm */
			CPK_CompleteDisp(cpk);
		} else {

			/* Framebuffer switching and waiting for the end of drawing */
			smp_DisplayFrameWait();
		}

		/* Pad control  */
		smp_PadCtrl(cpk);

		/* See if the movie has ended */
		if (CPK_GetPlayStatus(cpk) == CPK_STAT_PLAY_END) {

			/* Waiting to display the last frame  */
			/* Wait to switch frame buffer. */
			SCL_DisplayFrame();

			/* Destroy the Movie. */
			CPK_DestroyGfsMovie(cpk);

			/* Close the file.*/
			fileClose(gfs);

			restart = 1;
		}
	}
}
